import 'dart:async';
import 'dart:convert';

import 'package:analyzer/dart/analysis/context_builder.dart';
import 'package:analyzer/dart/analysis/context_locator.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/file_system/file_system.dart';
// ignore: implementation_imports
import 'package:analyzer/src/dart/analysis/driver.dart'
    show AnalysisDriver, AnalysisDriverGeneric;
// ignore: implementation_imports
import 'package:analyzer/src/dart/analysis/driver_based_analysis_context.dart';
import 'package:analyzer_plugin/plugin/plugin.dart';
import 'package:analyzer_plugin/protocol/protocol_generated.dart' as plugin;
import 'package:glob/glob.dart';
import 'package:logging/logging.dart';
import 'package:pub_semver/pub_semver.dart';
import 'package:yaml/yaml.dart';

import 'package:pt_linter/mirror_visitor.dart';
import 'package:pt_linter/utils/map_utils.dart';

final _logger = Logger('string_literal_finder.plugin');

class MirrorPlugin extends ServerPlugin {
  MirrorPlugin(ResourceProvider provider) : super(provider);

  static const excludedFolders = [
    '.dart_tool/',
    '/.symlinks/',
    '/.vendors/',
    '/example/',
    '/test/'];

  var _filesFromSetPriorityFilesRequest = <String>[];
  final Map<AnalysisDriverGeneric, AnalysisOptions> _options = {};

  @override
  List<String> get fileGlobsToAnalyze => <String>['**/*.dart'];

  @override
  String get name => 'pt_linter';

  @override
  String get version => '0.1.1';

  @override
  bool isCompatibleWith(Version serverVersion) => true;

  @override
  void contentChanged(String path) {
    super.driverForPath(path)?.addFile(path);
  }

  @override
  AnalysisDriverGeneric createAnalysisDriver(plugin.ContextRoot contextRoot) {
    final rootPath = contextRoot.root;
    final locator =
        ContextLocator(resourceProvider: resourceProvider).locateRoots(
      includedPaths: <String>[rootPath],
      excludedPaths: <String>[
        ...contextRoot.exclude,
      ],
      optionsFile: contextRoot.optionsFile,
    );

    if (locator.isEmpty) {
      final error = StateError('Unexpected empty context');
      channel.sendNotification(plugin.PluginErrorParams(
        true,
        error.message,
        error.stackTrace.toString(),
      ).toNotification());

      throw error;
    }

    final builder = ContextBuilder(
      resourceProvider: resourceProvider,
    );

    final analysisContext = builder.createContext(contextRoot: locator.first);
    // ignore: avoid_as
    final context = analysisContext as DriverBasedAnalysisContext;
    final dartDriver = context.driver;
    try {
      final analysisOptions =
          (_options[dartDriver] ??= _getAnalysisOptions(dartDriver));

      runZonedGuarded(
        () {
          dartDriver.results.listen((analysisResult) {
            if (analysisResult is ResolvedUnitResult) {
              _processResult(
                dartDriver,
                analysisOptions,
                analysisResult,
              );
            } else if (analysisResult is ErrorsResult) {
              _logger
                  .severe('analysis produce errors. ${analysisResult.errors}');
            }
          });
        },
        (Object e, StackTrace stackTrace) {
          channel.sendNotification(
            plugin.PluginErrorParams(
              false,
              'pt_linter. Unexpected error: ${e.toString()}',
              stackTrace.toString(),
            ).toNotification(),
          );
        },
      );
    } catch (e, stackTrace) {
      channel.sendNotification(
        plugin.PluginErrorParams(
          false,
          'pt_linter. Unexpected error: ${e.toString()}',
          stackTrace.toString(),
        ).toNotification(),
      );
    }

    return dartDriver;
  }

  AnalysisOptions _getAnalysisOptions(AnalysisDriver analysisDriver) {
    final optionsPath = analysisDriver.analysisContext?.contextRoot?.optionsFile;
    final exists = optionsPath?.exists ?? false;
    if (!exists || optionsPath == null) {
      _logger.warning('Unable to resolve optionsFile.');
      return AnalysisOptions(excludeGlobs: []);
    }
    return AnalysisOptions.loadFromYaml(optionsPath.readAsStringSync());
  }

  @override
  Future<plugin.AnalysisSetPriorityFilesResult> handleAnalysisSetPriorityFiles(
      plugin.AnalysisSetPriorityFilesParams parameters) async {
    _filesFromSetPriorityFilesRequest = parameters.files;
    _updatePriorityFiles();
    return plugin.AnalysisSetPriorityFilesResult();
  }

  void _updatePriorityFiles() {
    final filesToFullyResolve = {
      ..._filesFromSetPriorityFilesRequest,
      for (final driver2 in driverMap.values) ...(driver2 as AnalysisDriver).addedFiles,
    };
    final filesByDriver = <AnalysisDriverGeneric, List<String>>{};
    for (final file in filesToFullyResolve) {
      final contextRoot = contextRootContaining(file);
      if (contextRoot != null) {
        final driver = driverMap[contextRoot];
        filesByDriver.putIfAbsent(driver, () => <String>[]).add(file);
      }
    }
    filesByDriver.forEach((driver, files) => driver.priorityFiles = files);
  }

  void _processResult(
    AnalysisDriver dartDriver,
    AnalysisOptions analysisOptions,
    ResolvedUnitResult analysisResult,
     ) {
    try {

      bool shouldProcess = excludedFolders.every((element) => !analysisResult.path.contains(element));
      if(!shouldProcess) return;

      if (analysisResult.unit != null && analysisResult.libraryElement != null) {
        final mirrorChecker = MirrorChecker(analysisResult.unit);
        final issues = mirrorChecker.enumToStringErrors();
        

        if (issues.isNotEmpty) {
          channel.sendNotification(
            plugin.AnalysisErrorsParams(
              analysisResult.path,
              issues.keys.map((issue) => analysisErrorFor(analysisResult.path, issue, analysisResult.unit)).toList(),
            ).toNotification(),
          );
        } else {
          channel.sendNotification(plugin.AnalysisErrorsParams(analysisResult.path, []).toNotification());
        }
      } else {
        channel.sendNotification(plugin.AnalysisErrorsParams(analysisResult.path, []).toNotification());
      }
    } on Exception catch (e, stackTrace) {
      channel.sendNotification(plugin.PluginErrorParams(false, e.toString(), stackTrace.toString()).toNotification());
    }
  }
  @override
  Future<plugin.EditGetFixesResult> handleEditGetFixes(
      plugin.EditGetFixesParams parameters) async {
    try {
      // bool shouldProcess = excludedFolders.every((element) => parameters.file.contains(element));
      // if(!shouldProcess) return plugin.EditGetFixesResult([]);
      
      final driver = driverForPath(parameters.file) as AnalysisDriver;

      final analysisResult = await driver.getResult2(parameters.file);
      if (analysisResult is! ResolvedUnitResult) {
        return plugin.EditGetFixesResult([]);
      }

      final resolvedResult = analysisResult as ResolvedUnitResult;

      // Get errors and fixes for the file.
      final mirrorChecker = MirrorChecker(resolvedResult.unit);
      final results = mirrorChecker.enumToStringErrors();

      // Return any fixes that are for the expected file.
      final fixes = <plugin.AnalysisErrorFixes>[];
      for (var error in results.keys) {
        plugin.PrioritizedSourceChange fix = results[error];
        if (fix != null &&
            fix.change.edits.single.file == parameters.file &&
            fix.change.edits.single.edits.isNotEmpty) {
          fixes.add(plugin.AnalysisErrorFixes(analysisErrorFor(resolvedResult.path, error, resolvedResult.unit),
              fixes: [fix]));
        }
      }

      return plugin.EditGetFixesResult(fixes);
    } catch (e, stackTrace) {
      // Notify the analyzer that an exception happened.
      channel.sendNotification(plugin.PluginErrorParams(false, e.toString(), stackTrace.toString()).toNotification());
      return plugin.EditGetFixesResult([]);
    }
  }
}


class AnalysisOptions {
  AnalysisOptions({this.excludeGlobs});

  static AnalysisOptions loadFromYaml(String yamlSource) {
    final yaml =
        json.decode(json.encode(loadYaml(yamlSource))) as Map<String, dynamic>;
    final options = yaml['string_literal_finder'] as Map<String, dynamic>;
    final excludeGlobs = options['exclude_globs'] as List<dynamic>;
    if (options == null || excludeGlobs == null) {
      return AnalysisOptions(excludeGlobs: []);
    }
    return AnalysisOptions(
      excludeGlobs: excludeGlobs.cast<String>().map((e) => Glob(e)).toList(),
    );
  }

  final List<Glob> excludeGlobs;

  bool isExcluded(String path) {
    return excludeGlobs.any((glob) => glob.matches(path));
  }
}
